import json
from os.path import join, basename
from shutil import copy
from tempfile import TemporaryDirectory

from juq.renumber import renumber_cmd
from tests.utils import TEST_DIR


def test_renumber_execution_counts():
    name = 'test-renumber'
    in_path = join(TEST_DIR, f"{name}.ipynb")
    expected_path = join(TEST_DIR, f"{name}-out.ipynb")
    with TemporaryDirectory() as tmpdir:
        out_path = join(tmpdir, basename(expected_path))
        renumber_cmd.callback(in_path, out_path)
        with open(out_path, 'r') as f:
            actual = json.load(f)
    # with open(expected_path, 'w') as f:
    #     json.dump(actual, f, indent=1)
    with open(expected_path, 'r') as f:
        expected = json.load(f)
    assert actual == expected


def test_renumber_execution_counts_inplace():
    name = 'test-renumber'
    orig_path = join(TEST_DIR, f"{name}.ipynb")
    expected_path = join(TEST_DIR, f"{name}-out.ipynb")
    with TemporaryDirectory() as tmpdir:
        nb_path = join(tmpdir, basename(orig_path))
        copy(orig_path, nb_path)
        # out_path = join(tmpdir, basename(expected_path))
        renumber_cmd.callback(nb_path, in_place=True)
        with open(nb_path, 'r') as f:
            actual = json.load(f)
    # with open(expected_path, 'w') as f:
    #     json.dump(actual, f, indent=1)
    with open(expected_path, 'r') as f:
        expected = json.load(f)
    assert actual == expected



from __future__ import annotations

from utz import decos

from juq.cli import with_nb
from juq.papermill import papermill, nb_opts


def papermill_clean_cell(
    cell,
    keep_ids: bool = False,
    keep_tags: bool | None = False,
):
    if not keep_ids and 'id' in cell:
        del cell['id']
    if 'metadata' in cell:
        metadata = cell['metadata']
        for k in [ 'papermill', 'execution', 'widgets' ]:
            if k in metadata:
                del metadata[k]
        if keep_tags is True:
            if 'tags' not in metadata:
                metadata['tags'] = []
        elif keep_tags is False:
            if 'tags' in metadata and not metadata['tags']:
                del metadata['tags']
    return cell


def papermill_clean(
    nb,
    keep_ids: bool = False,
    keep_tags: bool | None = False,
):
    """Remove Papermill metadata from a notebook.

    Removes `.metadata.papermill` and `.cells[*].metadata.{papermill,execution,widgets}`.
    """
    nb['cells'] = [
        papermill_clean_cell(
            cell,
            keep_ids=keep_ids,
            keep_tags=keep_tags,
        )
        for cell in nb['cells']
    ]
    metadata = nb['metadata']
    if 'papermill' in metadata:
        del metadata['papermill']
    return nb


papermill_clean_cmd = decos(
    papermill.command('clean'),
    nb_opts,
    with_nb,
)(papermill_clean)



import json
from os.path import join
from subprocess import check_call, CalledProcessError
from tempfile import TemporaryDirectory

from papermill import execute_notebook, PapermillExecutionError
from pytest import raises
from utz import env

from juq.papermill.clean import papermill_clean_cmd
from juq.papermill.run import papermill_run_cmd
from tests.utils import MERGE_OUTPUTS_DIR, TEST_DIR


def test_papermill_clean():
    nb_path = join(MERGE_OUTPUTS_DIR, "merged-outputs.ipynb")
    with TemporaryDirectory() as tmpdir:
        post_papermill_path = join(tmpdir, 'post-papermill.ipynb')
        execute_notebook(nb_path, post_papermill_path)
        with open(post_papermill_path, 'r') as f:
            post_papermill = json.load(f)
        assert 'papermill' in post_papermill['cells'][0]['metadata']
        assert 'papermill' in post_papermill['metadata']

        cleaned_path = join(tmpdir, 'cleaned.ipynb')
        papermill_clean_cmd.callback(post_papermill_path, cleaned_path)
        with open(cleaned_path, 'r') as f:
            cleaned_nb = json.load(f)
        split_outputs_nb_path = join(MERGE_OUTPUTS_DIR, "split-outputs.ipynb")
        with open(split_outputs_nb_path, 'r') as f:
            split_outputs_nb = json.load(f)
    assert cleaned_nb == split_outputs_nb


def test_papermill_error():
    nb_path = join(TEST_DIR, "test-err.ipynb")
    with TemporaryDirectory() as tmpdir, env(NO_COLOR="1"):
        out_path1 = join(tmpdir, 'out1.ipynb')
        with raises(PapermillExecutionError):
            papermill_run_cmd.callback(
                nb_path, out_path1,
                keep_ids=False,
                keep_tags=None,
                parameter_strs=(),
                request_save_on_cell_execute=False,
                autosave_cell_every=False,
            )

        with open(join(TEST_DIR, "test-err-out.ipynb"), 'rt') as f:
            expected = f.read()

        with open(out_path1, 'rt') as f:
            actual1 = f.read()
        assert actual1 == expected

        out_path2 = join(tmpdir, 'out2.ipynb')
        with raises(CalledProcessError):
            check_call([ 'juq', 'papermill', 'run', '-n1', nb_path, out_path2 ])

        with open(out_path2, 'rt') as f:
            actual2 = f.read()
        assert actual2 == expected



from copy import deepcopy

from utz import err, decos

from juq.cli import with_nb, cli


def merge_cell_outputs(cell):
    if 'outputs' not in cell:
        return cell
    outputs = cell['outputs']
    new = []
    i = 0
    n = len(outputs)
    while i < n:
        cur = deepcopy(outputs[i])
        if cur['output_type'] == "stream" and cur.get('name'):
            name = cur['name']
            while True:
                i += 1
                if i == n:
                    break
                nxt = outputs[i]
                if nxt['output_type'] == "stream" and nxt.get('name') == name:
                    cur['text'] += nxt['text']
                else:
                    break
        else:
            i += 1
        new.append(cur)
    cell['outputs'] = new
    return cell


def merge_outputs(nb):
    """Merge consecutive "stream" outputs (e.g. stderr)."""
    nb['cells'] = [
        merge_cell_outputs(cell)
        for cell in nb['cells']
    ]
    return nb


merge_outputs_cmd = decos(
    cli.command('merge-outputs'),
    with_nb,
)(merge_outputs)



from click import option
from utz import decos

from juq.cli import cli


@cli.group
def papermill():
    """Wrapper for Papermill commands (`clean`, `run`)."""
    pass


nb_opts = decos(
    option('-I', '--keep-ids', is_flag=True, help='Keep cell ids; by default they are removed.'),
    option('-k/-K', '--keep-tags/--no-keep-tags', default=None, help='When a cell\'s `tags` array is empty, enforce its presence or absence in the output.'),
)



from __future__ import annotations

import json
from contextlib import nullcontext
from functools import wraps
from inspect import getfullargspec
from sys import stdin, stdout

from click import argument, group, option
from utz import recvs, call


@group()
def cli():
    pass


def with_nb_input(func):
    spec = getfullargspec(func)

    @wraps(func)
    @argument('nb_path', required=False)
    @argument('out_path', required=False)
    def wrapper(
        nb_path: str | None = None,
        out_path: str | None = None,
        **kwargs,
    ):
        if 'out_path' in kwargs and kwargs['out_path']:
            if out_path != kwargs['out_path']:
                raise ValueError(f"Specify -o/--out-path xor a 2nd positional arg: {out_path} != {kwargs['out_path']}")

        ctx = nullcontext(stdin) if nb_path == '-' or nb_path is None else open(nb_path, 'r')
        with ctx as f:
            nb_str = f.read()
            indent = kwargs.pop('indent', None)
            if indent is None:
                if nb_str.startswith('{'):
                    if nb_str[1] == "\n":
                        idx = 2
                        indent = 0
                        while nb_str[idx] == ' ':
                            idx += 1
                            indent += 1
                    else:
                        indent = None
                else:
                    raise ValueError(f"Cannot infer `indent` from non-JSON input beginning with {nb_str[:30]}")

            trailing_newline = kwargs.pop('trailing_newline', None)
            if trailing_newline is None:
                trailing_newline = nb_str.endswith('\n')
            nb = json.loads(nb_str)
        return call(
            func,
            **kwargs,
            nb_path=nb_path,
            out_path=out_path,
            nb=nb,
            indent=indent,
            trailing_newline=trailing_newline,
        )
    return wrapper


def with_nb(func):
    @option('-a', '--ensure-ascii', is_flag=True, help='Octal-escape non-ASCII characters in JSON output')
    @option('-i', '--in-place', is_flag=True, help='Modify [NB_PATH] in-place')
    @option('-n', '--indent', type=int, help='Indentation level for the output notebook JSON (default: infer from input)')
    @option('-o', '--out-path', help='Write to this file instead of stdout')
    @option('-t/-T', '--trailing-newline/--no-trailing-newline', default=None, help='Enforce presence or absence of a trailing newline (default: match input)')
    @with_nb_input
    @wraps(func)
    def wrapper(
        nb_path: str,
        *args,
        out_path: str | None = None,
        ensure_ascii: bool = False,
        in_place: bool = False,
        indent: int | None = None,
        trailing_newline: bool | None = None,
        **kwargs,
    ):
        """Merge consecutive "stream" outputs (e.g. stderr)."""
        if in_place:
            if out_path:
                raise ValueError("Cannot use `-i` with `-o`")
            if not nb_path or nb_path == '-':
                raise ValueError("Cannot use `-i` without explicit `nb_path`")
            out_path = nb_path

        kwargs['nb_path'] = nb_path
        kwargs['out_path'] = out_path
        rv = call(func, *args, **kwargs)
        if isinstance(rv, tuple):
            nb, exc = rv
        elif isinstance(rv, dict):
            nb = rv
            exc = None
        else:
            raise ValueError(f"Unrecognized with_nb return value {type(rv)}: {str(rv)[:100]}")

        out_ctx = nullcontext(stdout) if out_path == '-' or out_path is None else open(out_path, 'w')
        with out_ctx as f:
            json.dump(nb, f, indent=indent, ensure_ascii=ensure_ascii)
            if trailing_newline:
                f.write('\n')

        if exc:
            raise exc

    return wrapper



from __future__ import annotations

import json
from os.path import join, basename
from tempfile import TemporaryDirectory

from juq.papermill.run import papermill_run_cmd
from tests.test_papermill_clean import TEST_DIR


def check(
    in_name: str,
    expected_name: str | None = None,
    **kwargs,
):
    in_path = join(TEST_DIR, in_name)
    if expected_name is None:
        expected_name = in_name
    expected_path = join(TEST_DIR, expected_name)
    name = basename(expected_path)
    with TemporaryDirectory() as tmpdir:
        out_path = join(tmpdir, name)
        papermill_run_cmd.callback(in_path, out_path, **kwargs)
        with open(out_path, 'r') as f:
            actual = json.load(f)
    with open(expected_path, 'r') as f:
        expected = json.load(f)
    assert actual == expected


def test_mixed_tags():
    check("mixed-tags.ipynb")


def test_mixed_tags_params():
    check(
        "mixed-tags-params.ipynb",
        "mixed-tags-params-222.ipynb",
        parameter_strs=("num=222",),
    )


def test_run_previously_run_nb():
    check(
        "mixed-tags-params-222.ipynb",
        "mixed-tags-params-333.ipynb",
        parameter_strs=("num=333",),
    )


def test_mixed_tags_keep():
    check(
        "mixed-tags.ipynb",
        "mixed-tags-keep.ipynb",
        keep_tags=True,
    )


def test_mixed_tags_drop():
    check(
        "mixed-tags.ipynb",
        "mixed-tags-drop.ipynb",
        keep_tags=False,
    )



import json

from click import argument, option

from juq.cli import with_nb_input, cli

CELL_TYPE_ABBREVS = {
    'c': 'code',
    'm': 'markdown',
    'md': 'markdown',
    'r': 'raw',
}


@cli.command
@option('-m/-M', '--metadata/--no-metadata', default=None, help='Explicitly include or exclude each cell\'s "metadata" key. If only `-m` is passed, only the "metadata" value of each cell is printed')
@option('-o/-O', '--outputs/--no-outputs', default=None, help='Explicitly include or exclude each cell\'s "outputs" key. If only `-o` is passed, only the "outputs" value of each cell is printed')
@option('-s/-S', '--source/--no-source', default=None, help='Explicitly include or exclude each cell\'s "source" key. If only `-s` is passed, the source is printed directly (not as JSON)')
@option('-t', '--cell-type', help='Only print cells of this type. Recognizes abbreviations: "c" for "code", {"m","md"} for "markdown", "r" for "raw"')
@argument('cells_slice')
@with_nb_input
def cells(cell_type, cells_slice, nb, **flags):
    """Slice/Filter cells."""
    cells = nb['cells']
    if cell_type:
        if cell_type in CELL_TYPE_ABBREVS:
            cell_type = CELL_TYPE_ABBREVS[cell_type]
        cells = [
            cell
            for cell in cells
            if cell['cell_type'] == cell_type
        ]

    def slice_cell(cell):
        num_flags = sum(1 if v else 0 for v in flags.values())
        if num_flags > 1:
            return {
                k: cell[k]
                for k, v in flags.items()
                if v and k in cell
            }
        elif num_flags == 1:
            for k, v in flags.items():
                if v:
                    if k == 'source':
                        source = cell['source']
                        return ''.join(source) if isinstance(source, list) else source
                    else:
                        return cell[k]
        else:
            return {
                k: cell[k]
                for k, v in cell.items()
                if flags.get(k)is not False
            }

    pcs = cells_slice.split(':')
    if len(pcs) == 1:
        cell = cells[int(pcs[0])]
        cell = slice_cell(cell)
        obj = cell
    elif len(pcs) == 2:
        cells = cells[slice(*map(lambda x: int(x.strip()) if x.strip() else None, pcs))]
        cells = [
            slice_cell(cell)
            for cell in cells
        ]
        obj = cells
    else:
        raise ValueError(f"Unrecognized <cells slice>: {cells_slice}")

    if isinstance(obj, str):
        print(obj)
    else:
        print(json.dumps(obj, indent=2))



from click import option
from nbformat import NotebookNode
from utz import silent, err, decos

from juq.cli import cli, with_nb


def renumber(
    nb: NotebookNode,
    quiet: bool = False,
):
    """Renumber cells (and outputs) with non-null `execution_count` fields (beginning from 1).

    Simulates the executed cells having been executed in order (e.g. to avoid needing to re-run a whole notebook after
    rearranging cells that don't depend on one another).
    """
    log = silent if quiet else err
    nxt_idx = 1
    for cell_idx, cell in enumerate(nb["cells"]):
        found = False
        if cell.get('execution_count') is not None:
            log(f".cells[{cell_idx}].execution_count: {cell['execution_count']} → {nxt_idx}")
            cell['execution_count'] = nxt_idx
            found = True
        for output_idx, output in enumerate(cell.get("outputs", [])):
            if output.get('execution_count') is not None:
                log(f".cells[{cell_idx}].outputs[{output_idx}].execution_count: {output['execution_count']} → {nxt_idx}")
                output['execution_count'] = nxt_idx
                found = True
        if found:
            nxt_idx += 1
    return nb


renumber_cmd = decos(
    cli.command,
    option('-q', '--quiet', is_flag=True, help="Suppress logging info about each `execution_count` update to stderr"),
    with_nb,
)(renumber)



from __future__ import annotations

import json
from os.path import join
from sys import stdout
from tempfile import TemporaryDirectory
from typing import Tuple

from click import option
from papermill.cli import _resolve_type
from utz import decos, env

from juq.cli import with_nb
from juq.merge_outputs import merge_outputs
from juq.papermill import nb_opts, papermill
from juq.papermill.clean import papermill_clean


INJECTED_TAGS = { "papermill-error-cell-tag", 'injected-parameters' }


class AlignmentError(ValueError):
    """Indicates a failure to align a generated notebook against its input.

    Papermill always adds empty ``.cells[].metadata.tags`` arrays, which we try to undo when -k/--keep-tags is not
    specified. Output notebooks may have "injected-parameters" cells inserted by Papermill, so we attempt to align the
    input and output notebooks, to mirror presence/absence of ``tags``. ``AlignmentError`` is raised when this process
    fails (e.g. because there appear to be additional, unrecognized cells added to the "output" notebook).
    """
    pass


def harmonize_empty_tags(cells0, cells1, exc):
    """When -k/--keep-tags is not specified, remove empty ``.cells[].metadata.tags`` arrays added by Papermill."""
    idx1 = 0

    def skip_injected_param_cells():
        """Skip ``idx1`` past any "injected-parameters" or "error" cells Papermill may have inserted."""
        nonlocal idx1
        while True:
            cell1 = cells1[idx1]
            md1 = cell1.get('metadata', {})
            tags1 = md1.get('tags')
            if not tags1 or not INJECTED_TAGS & set(tags1):
                return cell1, md1, tags1
            idx1 += 1

    try:
        for idx0, cell0 in enumerate(cells0):
            md0 = cell0['metadata']
            tags0 = md0.get('tags')
            if tags0 and len(tags0) == 1 and tags0[0] in INJECTED_TAGS:
                cell1 = cells1[idx1]
                md1 = cell1.get('metadata', {})
                tags1 = md1.get('tags')
                if tags0 != tags1:
                    raise AlignmentError(f"{tags0=} != {tags1=}")
            else:
                cell1, md1, tags1 = skip_injected_param_cells()
                source0 = cell0.get('source')
                source1 = cell1.get('source')
                if source0 != source1:
                    raise AlignmentError(f"Cell {idx0=}: {source0=} != {source1=}")
                if 'tags' not in md0 and tags1 == []:
                    del md1['tags']
            idx1 += 1
        try:
            skip_injected_param_cells()
        except IndexError:
            pass
        if idx1 != len(cells1):
            raise AlignmentError(f'"After" notebook has {len(cells1) - idx1} extra cells: {json.dump(cells1[idx1:], stdout)}')
    except AlignmentError as align_exc:
        if exc:
            align_exc.__cause__ = exc
        exc = align_exc
    return exc


def papermill_run(
    nb_path,
    keep_ids: bool = False,
    keep_tags: bool | None = None,
    parameter_strs: Tuple[str, ...] = (),
    request_save_on_cell_execute: bool | None = None,
    autosave_cell_every: int | None = None,
):
    """Run a notebook using Papermill, clean nondeterministic metadata, normalize output streams."""
    from papermill import PapermillExecutionError, execute_notebook

    parameters = {}
    for param_str in parameter_strs:
        pcs = param_str.split('=', 1)
        if len(pcs) != 2:
            raise ValueError(f"Unrecognized parameter string: {param_str}")
        k, v = pcs
        parameters[k] = _resolve_type(v)

    exc = None
    with TemporaryDirectory() as tmpdir:
        tmp_out = join(tmpdir, 'out.ipynb')
        try:
            with env(PAPERMILL='1'):
                execute_notebook(
                    nb_path, tmp_out,
                    parameters=parameters,
                    request_save_on_cell_execute=request_save_on_cell_execute,
                    **({} if autosave_cell_every is None else dict(autosave_cell_every=autosave_cell_every)),
                )
        except PapermillExecutionError as e:
            exc = e

        with open(tmp_out, 'r') as f:
            nb = json.load(f)

    if keep_tags is None:
        with open(nb_path, 'r') as f:
            nb0 = json.load(f)
        cells0 = nb0['cells']
        cells1 = nb['cells']
        exc = harmonize_empty_tags(cells0, cells1, exc)

    nb = papermill_clean(nb, keep_ids=keep_ids, keep_tags=keep_tags)
    nb = merge_outputs(nb)
    return nb, exc


papermill_run_cmd = decos(
    papermill.command('run'),
    nb_opts,
    option('-p', '--parameter', 'parameter_strs', multiple=True, help='"<k>=<v>" variable to set, while executing the notebook'),
    option('-s', '--request-save-on-cell-execute', is_flag=True, default=None, help="Request save notebook after each cell execution"),
    option('-S', '--autosave-cell-every', type=int, help="How often in seconds to autosave the notebook during long cell executions (0 to disable)"),
    with_nb,
)(papermill_run)



import json
from os import path
from tempfile import TemporaryDirectory

from juq.merge_outputs import merge_outputs_cmd
from tests.test_papermill_clean import MERGE_OUTPUTS_DIR


def test_merge_cell_outputs():
    nb_path = path.join(MERGE_OUTPUTS_DIR, "split-outputs.ipynb")
    out_name = "merged-outputs.ipynb"
    expected_path = path.join(MERGE_OUTPUTS_DIR, out_name)
    with TemporaryDirectory() as tmpdir:
        actual_path = path.join(tmpdir, out_name)
        merge_outputs_cmd.callback(nb_path, actual_path)
        with open(expected_path, 'r') as f:
            expected_nb = json.load(f)
        with open(actual_path, 'r') as f:
            actual_nb = json.load(f)
    assert expected_nb == actual_nb
